﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.ObjectModel;
using System.Data.EntityClient;
using System.Data.SqlClient;
using DeepViewDataModel.DataModel;
using DeepViewDataModel.Common;
using System.Collections.Concurrent;
using System.IO;
using DZConverterLib.DAL;
using DZConverterLib.Auth;
using pdftron;
using System.Data.Objects.DataClasses;
using pdftron.PDF;
using System.Drawing;
using System.Drawing.Imaging;
using pdftron.SDF;
using Microsoft.DeepZoomTools;
using DeepView.Toolkit.GuidGenerator;
using System.Threading;
using System.Net.Cache;
using System.Net;
using LimitTaskScheduler;
using System.Threading.Tasks;
using System.Runtime.Caching;

namespace DZConverterLib
{
    public class DZConverter: IDisposable
    {
        //public const string ProjectRoot = @"D:\DeepViewConverter\";
        //public const string DownloadDir_PDF = ProjectRoot + @"\PDFPath";
        public string DownloadDir_PDF;
        private List<String> DeepZoopDest = null;
        private EntityCollection<DocPage> pageSet = null;
        public int pageCount = 0;
        public ConverterPDF ConverterItem;
        public event EventHandler<MakeConverterErrorArgs> ConverterMakeError;
        public event EventHandler<EndConverterEventArgs> EndConverter;

        public DZConverter()
        {
            pdftron.PDFNet.Initialize("daelim.co.kr(Doftech Corp):CPU:2::W:AMC(20120315):EF6E886F25A414FFB5F8C1F2999CF2DA33DC6C5164315BAF7011B87AF0FA");
            DeepZoopDest = new List<string>();
            pageSet = new EntityCollection<DocPage>();
        }

        private void SendError(int Level,string ErrorCode, string _Exception, bool ThreadStop)
        {
            if (ConverterMakeError != null)
                ConverterMakeError(this, new MakeConverterErrorArgs 
                                            {
                                                ThreadStop = ThreadStop, 
                                                ConverterID = ConverterItem.ID,
                                                _Exception = _Exception,
                                                ErrorCode = ErrorCode,
                                                Level = Level
                                            });
        }

        public void GetSingleFile(object obj)
        {
            try
            {
                ConverterItem = obj as ConverterPDF;

                using (DeepViewEntities _deepview = new DeepViewEntities(ConnectStringBuilder.DeepViewConnectionString().ToString()))
                {
                    var _property = _deepview.Properties.Where(pro => pro.Property == ConverterItem.ProjectNo);

                    if (_property.Count() > 0)
                    {
                        DownloadDir_PDF = _property.Where(t => t.Type == PropertiesType.PropertiesType.Const_TileSorceStorage).First().Value;

                        if (!Directory.Exists(DownloadDir_PDF)) // 폴더 없을 시 생성
                        {
                            Directory.CreateDirectory(DownloadDir_PDF);
                        }
                    }
                }

                DownloadFile();
            }
            catch (Exception e)
            {
                SendError(0, "Converter Error ", e.ToString(), true);
            }
        }

        private void SetState(IConverterPDF.ConverterStatus Status)
        {
            using (DeepViewEntities _entity = new DeepViewEntities(ConnectStringBuilder.DeepViewConnectionString().ToString()))
            {
                var _items = _entity.ConverterPDF.Where(converter => converter.ID == this.ConverterItem.ID);
                if (_items.Count() > 0)
                {
                    if (Status == IConverterPDF.ConverterStatus.Create)
                        _items.First().StartDateTime = DateTime.Now;

                    _items.First().Status = (int)Status;
                    _entity.SaveChanges();
                }
            }  
        }

        private void SetCurrentPage(int CurrentPage)
        {
            using (DeepViewEntities _entity = new DeepViewEntities(ConnectStringBuilder.DeepViewConnectionString().ToString()))
            {
                var _items = _entity.ConverterPDF.Where(converter => converter.ID == this.ConverterItem.ID);
                if (_items.Count() > 0)
                {
                    _items.First().CurrentPage = CurrentPage;
                    _entity.SaveChanges();
                }
            }
        }

        private static object _lock = new object();
     
        public void DownloadFile()
        {
            string DocUri = ConverterItem.PdfUrl; //PDF 전체 경로
            string FileName = DocUri.Remove(0, DocUri.LastIndexOf("/") + 1);
            string ProjectFolderPath = DownloadDir_PDF + "\\" + ConverterItem.ProjectNo + "_Tile"; //프로젝트 폴더
            string ItemListPath = ProjectFolderPath + "\\" + (System.Convert.ToInt64(ConverterItem.SharepointItemID) / 100).ToString();
            string ItemPath = ItemListPath + "\\" + ConverterItem.SharepointItemID;
            string DownloadFilePath = ItemPath + "\\" + FileName;
          
            #region 폴더 생성
            try
            {
                SetState(IConverterPDF.ConverterStatus.Create);

                DirectoryInfo _dir = new DirectoryInfo(ItemPath);

                if (_dir.Exists)
                {
                    _dir.Empty();
                }
                else
                {
                    Directory.CreateDirectory(ItemPath);
                }

            }
            catch (Exception ex)
            {
                //DeleteDirectoryDirs(ItemPath);
                SendError(1,"폴더삭제 에러",ex.ToString(), false);
            }

            #endregion

            #region 다운로드

            try
            {
                System.Net.WebClient _fileFactory = new System.Net.WebClient();
                _fileFactory.UseDefaultCredentials = true;
                _fileFactory.DownloadFile(new Uri(DocUri), DownloadFilePath);

            #endregion
            }
            catch (Exception ex)
            {
                SendError(1,"File Download Error : Url - " + DocUri + " Download Path : " + DownloadFilePath, ex.ToString(), true);
            }
           

            if (File.Exists(DownloadFilePath))
            {
                ConvertProcess(DownloadFilePath, ItemPath, FileName);
            }
            else
            {
                SendError(1,"File Not Found","if (File.Exists(DownloadFilePath))", true);
            }

        }

        public void ConvertProcess(string DownloadFilePath, string ItemPath, string FileName)
        {
            #region 기본 설정

            #endregion

            ResultSet Result_Check;
            #region PDF 체크
            try
            {
                Result_Check = Check_PDF(DownloadFilePath);
            }
            catch (Exception ex)
            {
                System.Diagnostics.Debug.WriteLine(ex.ToString());
                throw;
            }
            
            if (!Result_Check.result)
            {
                SendError(1, "PDF 체크 실패 : 보안설정으로 인한 PDF 체크 실패", Result_Check.Message, false);
            } 
            #endregion

            #region PDF에서 이미지로 변환

            try
            {
                if (!ConvertImage(ItemPath, FileName))
                {
                    SendError(0, "converter Check", "code if (!ConvertImage(ItemPath, FileName))", true);
                    return;
                } 
            }
            catch (Exception ex)
            {
                SendError(0, "converter Check ","if (!ConvertImage(ItemPath, FileName)) " + ex.ToString(), true);
                throw;
            }

            #endregion

            SetState(IConverterPDF.ConverterStatus.Crop);

            #region 이미지에서 딥줌 이미지로
            var _DeepZoomImageresult = ConvertDeepZoomImage(DeepZoopDest);

            int _resultCount = _DeepZoomImageresult.Where(f => f.Count() == 0).Count();

            if (_resultCount == DeepZoopDest.Count())
            {
                SetCurrentPage(_resultCount);
            }
            else
            {
                SendError(0, "converter Check Error", string.Join("\r\n", _DeepZoomImageresult), true);
                return;
            }

            #endregion

            #region 데이터베이스 입력
            var _infotoDBResult = SetPageInfoToDB();
            if (_infotoDBResult != null)
            {
                SendError(0, "SetPageInfoToDB()", _infotoDBResult, true);
                return;
            }
            #endregion

            //// virtual dir생성은 로컬에서만 가능
            //try
            //{
            //    VirtualDirCreate.CreateVirtualDir("comment.daelim.com", "MACHINE/WEBROOT/APPHOST/DeepView", prjNo + "_Tile", DownloadDir_PDF + "\\" + prjNo + "_Tile");
            //}
            //catch (Exception ex)
            //{
            //    SendError("VirtualDirCreate Error");
            //}

            if (EndConverter != null)
                EndConverter(this, new EndConverterEventArgs { ConverterPDF = ConverterItem});
        }

        public ResultSet Check_PDF(string filePath)
        {
            pdftron.PDF.PDFDoc doc = new pdftron.PDF.PDFDoc(filePath);
            #region 접근권한 체크
            if (!doc.InitSecurityHandler())
            {
                PDFNet.Terminate();
                return new ResultSet { result = false, Message = "Document authentication Error" };
            } 
            #endregion

            #region 암호화 여부
            if (doc.IsEncrypted())
            {
                PDFNet.Terminate();
                return new ResultSet { result = false, Message = "Document is Encrypted" };
            } 
            #endregion            
            return new ResultSet { result = true };
        }


        #region private Method
        private bool AuthUser2(Filefactory fileFactory, string PDFurl)
        {
            try
            {
                System.Net.HttpWebRequest request = (System.Net.HttpWebRequest)System.Net.WebRequest.Create(PDFurl);
                System.Net.NetworkCredential credentials = Crediential.GetAdminCrediential(PDFurl.ToString());
                request.Credentials = credentials;
                System.Net.HttpWebResponse response = (System.Net.HttpWebResponse)request.GetResponse();
                response.Close();
                fileFactory.Credentials = credentials;
            }
            catch (Exception ex)
            {
                System.Diagnostics.Debug.WriteLine(ex.ToString());
                return false;
            }

            return true;
        }
        //private System.Net.ICredentials AuthUser(string PDFurl)
        //{
        //    //System.Net.NetworkCredential credentials = null;
        //    //try
        //    //{
        //    //    var _url = new UriBuilder(new Uri(PDFurl).AbsoluteUri.Replace(new Uri(PDFurl).PathAndQuery, "/"));
        //    //    var _dns = _url.Host.Split('.');

        //    //    if (_dns.Count() == 3)
        //    //        if (_dns[0].ToLower() != "www")
        //    //        {
        //    //            _dns[0] = "www";
        //    //            _url.Host = _dns[0] + "." + _dns[1] + "." + _dns[2];
        //    //        }

        //    //    System.Net.HttpWebRequest request = (System.Net.HttpWebRequest)System.Net.WebRequest.Create(_url.Uri);
        //    //    credentials =
        //    //    //request.Credentials = CredentialCache.DefaultCredentials;
        //    //    System.Net.HttpWebResponse response = (System.Net.HttpWebResponse)request.GetResponse();   
        //    //    response.Close();
        //    }
        //    catch (Exception ex)
        //    {
        //        return credentials;
        //    }

        //    return credentials;
        //}

        private ImageCodecInfo GetEncoderInfo(String mimeType)
        {
            int j;
            ImageCodecInfo[] encoders;
            encoders = ImageCodecInfo.GetImageEncoders();
            for (j = 0; j < encoders.Length; ++j)
            {
                if (encoders[j].MimeType == mimeType)
                    return encoders[j];
            }
            return null;
        }

        private bool ConvertImage(string FilePath, string fileName)
        {
            pageSet = new EntityCollection<DocPage>();
            using (PDFDoc doc = new PDFDoc(FilePath + "\\" + fileName))
            {
                pdftron.PDFNet.Initialize("daelim.co.kr(Doftech Corp):CPU:2::W:AMC(20120315):EF6E886F25A414FFB5F8C1F2999CF2DA33DC6C5164315BAF7011B87AF0FA");
                Bitmap newBmp_;
                ImageCodecInfo DefaultImageCodecInfo = GetEncoderInfo("image/png");
                EncoderParameters DefaultEncoderParameters = new EncoderParameters(2);
                System.Drawing.Imaging.Encoder QualityEncoder = System.Drawing.Imaging.Encoder.Quality;
                System.Drawing.Imaging.Encoder ColorDepthEncoder = System.Drawing.Imaging.Encoder.ColorDepth;
                DefaultEncoderParameters.Param[0] = new EncoderParameter(QualityEncoder, 100L);
                DefaultEncoderParameters.Param[1] = new EncoderParameter(ColorDepthEncoder, 8L);
                DeepZoopDest = new List<string>();
                pageCount = doc.GetPageCount();

                using (DeepViewEntities _entity = new DeepViewEntities(ConnectStringBuilder.DeepViewConnectionString().ToString()))
                {
                    var _items = _entity.ConverterPDF.Where(converter => converter.ID == this.ConverterItem.ID);
                    if (_items.Count() > 0)
                    {
                        _items.First().TotalPages = pageCount;
                        _entity.SaveChanges();
                    }
                }

                #region 이미지 만들기

                for (int i = 1; i < pageCount + 1; i++)
                {
                    try
                    {
                        //SetCurrentPage(i);
                        using (pdftron.PDF.PDFDraw draw = new pdftron.PDF.PDFDraw())
                        {
                            ElementBuilder bld = new ElementBuilder();
                            ElementWriter writer = new ElementWriter();

                            var widthData = (int)doc.GetPage(i).GetPageWidth();
                            int heightData = (int)doc.GetPage(i).GetPageHeight();
                            pageSet.Add(new DocPage 
                                        {
                                            Id=  GuidGenerator.GetUniqueGuid() , 
                                            PageWidth = (widthData*3).ToString(), 
                                            PageHeight = (heightData*3).ToString(), 
                                            PageNumber = i, 
                                            PageAngle = 0,
                                        });

                            var rotation = doc.GetPage(i).GetRotation();
                            draw.SetImageSize((int)(widthData * 3), (int)(heightData * 3), true);
                            //draw.SetImageSize((int)(widthData), (int)(heightData), true);
                            draw.SetAntiAliasing(true);
                            draw.SetImageSmoothing(true);
                            newBmp_ = draw.GetBitmap(doc.GetPage(i));

                            using (MemoryStream _savestream = new MemoryStream())
                            {

                                //newBmp_.Save(@"c:\test.jpg");
                                newBmp_.Save(_savestream, DefaultImageCodecInfo, DefaultEncoderParameters);
                                //newBmp_.Save(_savestream, ImageFormat.Png);
                                newBmp_ = new Bitmap(_savestream);
                                ObjSet objset = new ObjSet();
                                Obj jbig2_hint = objset.CreateName("png");

                                string pagePath = FilePath + "\\" + i;

                                Directory.CreateDirectory(pagePath);
                                DeepZoopDest.Add(pagePath + "\\" + i + ".jpg");
                                newBmp_.Save(pagePath + "\\" + i + ".jpg");
                                newBmp_.Dispose();

                            }

                            GC.Collect();
                            GC.WaitForPendingFinalizers();
                        }

                        if (i == pageCount || i % 10 == 0)
                        {
                            int _skipCount = 0;

                            if (pageCount < 10 || i < 10)
                                _skipCount = 0;
                            else
                                _skipCount = i - 9; // 1페이지부터 가져온다.

                            ConvertDeepZoomImage(DeepZoopDest.Skip(_skipCount).Take(i));

                        }
                    }
                    catch (Exception ex)
                    {
                        SendError(0,"ConverterImage Error : Page NO -" + i.ToString(), ex.ToString(), true);
                        return false;
                    }
                }
                #endregion
                return true;
            }
        }

        private List<Task> DeepTask = new List<Task>();

        private List<string> ConvertDeepZoomImage(IEnumerable<string> FileSet)
        {
            Task<List<string>> _task = new Task<List<string>>(() =>
            {
                var _re = (from file in FileSet.Select((s, i) => new { Inx = i, file = s }).AsParallel().WithDegreeOfParallelism(10)
                           let result = ConvertImgToDeep(file.file, file.Inx)
                           select result).ToList();
                
                return _re;
                //return string.Join(",", _result);
            });
            _task.Start();
            _task.Wait();
            
            return _task.Result;
        }

        private string ConvertImgToDeep(string convertFileName, int Index)
        {
            string _result = string.Empty;

            ImageCreator creator = new ImageCreator
            {
                TileSize = 256,
                TileFormat = Microsoft.DeepZoomTools.ImageFormat.AutoSelect,
                CopyMetadata = false,
                ConversionCopyMetadata = false,
            };


            try
            {
                FileInfo convertFile = new FileInfo(convertFileName);

                if (convertFile.Exists)
                {

                    //var _dir = convertFile.Directory.CreateSubdirectory(Index.ToString());
                    //convertFile = convertFile.CopyTo(_dir.FullName + "\\" + convertFile.Name, true);
                    creator.Create(convertFile.FullName, convertFile.FullName.ToString().Replace(".jpg", ".xml"));
                    File.Delete(convertFile.FullName.ToString().Replace(".jpg", ".xml"));

                    if(convertFile.Directory.GetDirectories().Count() > 0)
                        convertFile.Directory.GetDirectories().First().GetDirectories().Where(f => f.Name.Length == 1)
                            .Where(d => System.Convert.ToInt16(d.Name) <= 7)
                            .ToList().ForEach(sub =>
                            {
                                sub.Empty();
                                sub.Delete();
                            });

                    convertFile.Delete();
                    SetCurrentPage(Index+1);
                }
            }
            catch (Exception ex)
            {
                System.Diagnostics.Debug.WriteLine(ex.ToString());

                if (ex.GetType() == typeof(System.ArgumentException))
                    _result = "DPI Exception. Page No :" + Index.ToString();
                else
                    _result = ex.Message;
            }

            return _result;
        }

        private string ConvertDeepZoomImage_old(List<String> FileSet)
        {
            ImageCreator crator = new ImageCreator();
            crator.TileSize = 256;
            crator.TileFormat = Microsoft.DeepZoomTools.ImageFormat.AutoSelect;
            crator.CopyMetadata = false;
            crator.ConversionCopyMetadata = false;
            try
            {
                for (int i = 0; i < FileSet.Count; i++)
                {
                    SetCurrentPage(i + 1);

                    if (File.Exists(FileSet[i]))
                    {
                        crator.Create(FileSet[i], FileSet[i].ToString().Replace(".jpg", ".xml"));
                        File.Delete(FileSet[i]);
                    }
                }
            }
            catch (Exception ex)
            {
                System.Diagnostics.Debug.WriteLine(ex.ToString());
                return ex.ToString();
            }
              
            return null;
        }

        private string SetPageInfoToDB()
        {
            try
            {
                #region 다 쓴 이미지 삭제
                //foreach (var item in DeepZoopDest)
                //{
                //    try
                //    {
                //        if (File.Exists(item.ToString()))
                //        {
                //            File.Delete(item.ToString());
                //        }
                //    }
                //    catch (Exception e)
                //    {
                //        /// 서버로 보내는 에러메시지 필요
                //    }

                //} 
                #endregion
                #region DB에 페이지 정보 입력
           

                using (CI_Entities deepViewEntity = new CI_Entities(ConnectStringBuilder.ProjectCIConnectString(ConverterItem.ProjectNo).ToString()))
                {
                    var _docinfo = deepViewEntity
                                    .DocInfo.Where(info => info.OrginPDFUrl == ConverterItem.PdfUrl);

                    if (_docinfo.Count() > 0)
                    {
                        System.Diagnostics.Stopwatch _watch = new System.Diagnostics.Stopwatch();
                        _watch.Start();
                        var _doc = _docinfo.First();
                        _doc.SharepointItemID = ConverterItem.SharepointItemID;
                        _doc.PageCount = pageCount;

                        pageSet.AsParallel().ForAll(p => p.DocInfo_Id = _doc.Id);

                        var _exceptDocInfo = (from newinfo in pageSet
                                              select new { newinfo.PageNumber, newinfo.DocInfo_Id,newinfo.PageWidth,newinfo.PageHeight})
                                                  .Except
                                                        (from oldinfo in _doc.DocPage
                                                         select new { oldinfo.PageNumber, oldinfo.DocInfo_Id, oldinfo.PageWidth, oldinfo.PageHeight });

                        var _delDocInfo = (from oldinfo in _doc.DocPage
                                           select new { oldinfo.PageNumber, oldinfo.DocInfo_Id, oldinfo.PageWidth, oldinfo.PageHeight })
                                                  .Except
                                                        (from newinfo in pageSet
                                                         select new { newinfo.PageNumber, newinfo.DocInfo_Id, newinfo.PageWidth, newinfo.PageHeight })
                                                         .ToList(); /// 삭제되는 collection은 하나씩 제거 될때마다 수정되기 때문에
                                                                    ///  foreach가 안된다. 그러므로 tolist로 가져와서 작업한다.
                        
                        //using (DeepView_SystemEntities SystemEntity = new DeepView_SystemEntities(ConnectStringBuilder.DeepViewSystemConnectionString().ToString()))
                        //{
                        //    SystemEntity.RemoveBackup_DocInfo.AddObject(new RemoveBackup_DocInfo
                        //    {
                        //        BackupUserID = "ConvertStation",
                        //        BackupTime= DateTime.Now,
                        //        ID = GuidGenerator.GetUniqueGuid(),
                        //        OriginDocInfoID = _doc.Id,
                        //        SharepointItemID = _doc.SharepointItemID,
                        //        OriginPDFUrl= _doc.OrginPDFUrl,
                        //        PageCount= _doc.PageCount,
                        //        ProjectNo = ConverterItem.ProjectNo,
                        //        Status = true
                        //    });


                        //    foreach (var item in _delDocInfo)
                        //    {
                        //        var _delitem = _doc.DocPage.Where(p => p.PageNumber == item.PageNumber).First();

                        //        deepViewEntity.DocPage.DeleteObject(_delitem);
                        //        //deepViewEntity.DocPage.DeleteObject(item);
                        //    }
                        //}
                        foreach (var item in _delDocInfo)
                        {
                            var _delitem = _doc.DocPage.Where(p => p.PageNumber == item.PageNumber).First();

                            deepViewEntity.DocPage.DeleteObject(_delitem);
                            //deepViewEntity.DocPage.DeleteObject(item);
                        }

                        deepViewEntity.SaveChanges();

                        foreach (var exceptitem in _exceptDocInfo)
                        {
                            var _lstPage = deepViewEntity.DocPage.Where(page => page.PageNumber == exceptitem.PageNumber
                                                                                    && page.DocInfo_Id == exceptitem.DocInfo_Id);

                            if (_lstPage.Count() > 0)
                            {
                                var _page = _lstPage.First();
                                _page.PageHeight = exceptitem.PageHeight;
                                _page.PageWidth = exceptitem.PageWidth;
                            }
                            else
                            {
                                _doc.DocPage.Add(pageSet.Where(p=>p.PageNumber == exceptitem.PageNumber).First());
                            }
                        }

                        deepViewEntity.SaveChanges();

                        _watch.Stop();
                        System.Diagnostics.Debug.WriteLine(_watch.ElapsedMilliseconds);
                    }
                    else
                    {

                        Guid guid = GuidGenerator.GetUniqueGuid();
                        DocInfo instace = new DocInfo
                        {
                            Id = guid,
                            OrginPDFUrl = ConverterItem.PdfUrl,
                            SharepointItemID = ConverterItem.SharepointItemID,
                            PageCount = pageCount,
                            DocPage = pageSet
                        };
                        deepViewEntity.DocInfo.AddObject(instace);
                    }

                    deepViewEntity.SaveChanges();
                }
            }
            catch (Exception e)
            {
                return e.ToString();
            }
               
            #endregion

            return null;
        }
    #endregion

        public void Dispose()
        {
            pdftron.PDFNet.Terminate();
            GC.Collect();
            GC.SuppressFinalize(this);
        }
     
    }

    public enum FinalStatus
    {
        /// <summary>
        /// Final PDF를 만들기 위한 순서에 추가
        /// </summary>
        Insert = 0,

        /// <summary>
        /// 앞의 final이 처리중일때 대기
        /// </summary>
        Wait = 1,

        /// <summary>
        /// 만들고 있는중
        /// 이때 데이터베이스에 Current Page는 만들고 있는 페이지
        /// </summary>
        Create = 2,

        /// <summary>
        /// final pdf 완료시
        /// </summary>
        Seccess = 99,

        /// <summary>
        /// 에러
        /// </summary>
        Error = 88
    }

}
